// Imperative and Object-Oriented Programming Methodology
// Uppsala University - 2018 Autumn
// Copyright © 2018 Henrik Schulze & Tobias Wrigstad
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include "interned.h"
#define NR_BUCKETS 17
typedef struct entry entry_t;

struct entry {
  char* string;
  int refcount;
  entry_t* next;
};

/// Global variable..
static entry_t** buckets; // 'strings' i uppgiftstextens 2.1.4.

/// Calculate hash for a string -- no need to read this function
static unsigned long string_hash(char* str) {
  unsigned long result = 0;
  do {
    result = result * 31 + *str;
  }
  while (*++str != '\0');
  return result;
}

/// Helper function to create an entry in the hash table. Probably
/// no need to change it, but allowed.
static entry_t* entry_create(char* str, entry_t* next) {
  entry_t* result = malloc(sizeof(entry_t));
  *result = (entry_t) {
    .string = strdup(str),
    .refcount = 1,
    .next = next
  };
  return result;
}

/// Helper function to destroy an entry in the hash table.
static void entry_destroy(entry_t* node) {
  //printf(" - entry_destroy! - %s -\n",node->string);
  free(node->string);
  free(node);
}

/// Returns the interned string for str. If str is already
/// interned, increase refcount by 1, else set initial refcount to
/// 1.
char* intstr_create(char* str) {
  /// Get the hash code for the string and use it to find the
  /// right bucket:
  unsigned long bucket = string_hash(str) % NR_BUCKETS;
  entry_t* entry = buckets[bucket];
  /// Search through the bucket for a string:
  while (entry != NULL) {
    /// Use strcmp here -- because we expect str to be not interned
    if (strcmp(str, entry->string) == 0) {
      /// If the string was already interned, return the
      /// interned string and increase its refcount by 1
      entry->refcount++;
      return entry->string;
    }
    entry = entry->next;
  }
  /// First time called on str -- create a new entry, add it to
  /// the front of the bucket, and return the entry's string:
  entry_t* new_entry = entry_create(str, buckets[bucket]);
  buckets[bucket] = new_entry;
  return new_entry->string;
}

/// Returns true if str is interned, else false
bool intstr_is_interned(char* str) {
  /// TODO
  bool is_interned = false;
  if (str != NULL) {
    unsigned long bucket = string_hash(str) % NR_BUCKETS;
    entry_t* cursor = buckets[bucket];
    while (cursor != NULL) {
      if (strcmp(str, cursor->string) == 0) {
        is_interned = true;
        cursor->next = NULL;
      }
      cursor = cursor->next;
    }
  }
  return is_interned;
}

/// Decrease the refcount of str by 1. If refcount hits 0, remove
/// the entry for the interned string and the interned string from
/// memory.
void intstr_destroy(char* str) {
  /// TODO
  if (str != NULL) {
    unsigned long bucket = string_hash(str) % NR_BUCKETS;
    entry_t** ptr_to_previous = &buckets[bucket];
    entry_t* cursor = NULL;
    bool found = false;
    while (*ptr_to_previous != NULL && !found) {
      if (str == (*ptr_to_previous)->string) {
        (*ptr_to_previous)->refcount--;
        if ((*ptr_to_previous)->refcount == 0) {
          cursor = (*ptr_to_previous)->next;
          entry_destroy(*ptr_to_previous);
          *ptr_to_previous = cursor;
        }
        found = true;
      }
      ptr_to_previous = &cursor;
    }
  }
}

/// Return the refcount for str if str is interned, else 0.
int intstr_refcount(char* str)
{
  if (str == NULL) return 0;
  unsigned long bucket = string_hash(str) % NR_BUCKETS;
  entry_t* entry = buckets[bucket];
  while (entry != NULL)
    {
      if (str == entry->string)
        {
          return entry->refcount;
        }
      else
        {
          entry = entry->next;
        }
    }
  return 0;
}

void intstr_init() {
  /// Initiate the global array of buckets:
  buckets = calloc(NR_BUCKETS, sizeof(entry_t* ));
}

void intstr_done() {
  /// TODO
  //entry_t* cursor = NULL;
  entry_t* next = NULL;
  for (int i = 0; i < NR_BUCKETS; ++i) {
    entry_t* cursor = buckets[i];
    while (cursor != NULL) {
      next = cursor->next;
      entry_destroy(cursor);
      cursor = next;
    }
  }
  free(buckets);
}
