// Imperative and Object-Oriented Programming Methodology
// Uppsala University - 2018 Autumn
// Copyright © 2018 Henrik Schulze & Tobias Wrigstad
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include "interned.h"

#define BUCKETS 17  

typedef struct entry entry_t;

struct entry
{
  char *string;
  int refcount; 
  entry_t* next;
};

/// Global variable..
static entry_t** buckets; /// Denna kallas "strings" i 2.1.4 i uppgiftstexten.

/// Calculate hash for a string -- no need to read this function
static unsigned long string_hash(char *str)
{
  unsigned long result = 0;
  do
    {
      result = result * 31 + *str;
    }
  while (*++str != '\0'); 
  return result;
}

/// Helper function to create an entry in the hash table. Probably
/// no need to change it, but allowed.
static entry_t* entry_create(char *str, entry_t* next)
{
  entry_t* result = calloc(1,sizeof(entry_t));
  *result = (entry_t) {
    .string = strdup(str),
    .refcount = 1,
    .next = next
  };
  return result;
}

/// Returns the interned string for str. If str is already
/// interned, increase refcount by 1, else set initial refcount to
/// 1.
char *intstr_create(char *str)
{
/// Get the hash code for the string and use it to find the right bucket:
  unsigned long bucket = string_hash(str) % BUCKETS;
  entry_t* entry = buckets[bucket];

  /// Search through the bucket for str:
  while (entry != NULL)
    {
      /// Use strcmp here -- because we expect str to be not interned:
      if (strcmp(str, entry->string) == 0)
        {
    /// If the string was already interned, increase its refcount by 1.
    /// Then return the interned string:
          ++entry->refcount;
          return entry->string;
        }
      else
        {
          entry = entry->next;
        }
    }

  /// First time called on str -- create a new entry, add it to
  /// the front of the bucket, and return the entry's string
  entry_t* new_entry = entry_create(str, buckets[bucket]);
  buckets[bucket] = new_entry;

  return new_entry->string;
}

/// Returns true if str is interned, else false
bool intstr_is_interned(char *str)
{
  /// TODO
  if (str == NULL) return false;
/// Get the hash code for the string and use it to find the right bucket:
  unsigned long bucket = string_hash(str) % BUCKETS;
  entry_t* cursor = buckets[bucket];
  /// Search through the bucket for str:
  while (cursor != NULL)
    {
      if (strcmp(str, cursor->string) == 0)
        {
          return true;
        }
      else
        {
          cursor = cursor->next;
        }
    }
  return false; /// -- str was not found
}

/// Decrease the refcount of str by 1. If refcount hits 0, remove the
/// entry for the interned string and the interned string from memory:
void intstr_destroy(char *str)
{
  /// TODO
  if (str == NULL) return;
/// Get the hash code for the string and use it to find the right bucket:
  unsigned long bucket = string_hash(str) % BUCKETS;
  entry_t* *cursor = &buckets[bucket];

  /// Search through the bucket for str:
  while (*cursor != NULL)
    {
      if (str == (*cursor)->string)
        {
    /// If the string was already interned, decrease its refcount by 1.
    /// If refcount then became zero, remove the entry:
          --((*cursor)->refcount);
          if ((*cursor)->refcount == 0)
            {
              entry_t* next = (*cursor)->next;
              free((*cursor)->string);
              free((*cursor));
              (*cursor) = next;
            }
          return;
        }
      else
        {
          cursor = &((*cursor)->next);
        }
    }
}

/// Return the refcount for str if str is interned, else 0.
int intstr_refcount(char *str)
{
  if (str == NULL) return 0;
  unsigned long bucket = string_hash(str) % BUCKETS;
  entry_t* *entry = &buckets[bucket];
  while ((*entry) != NULL)
    {
      if (str == (*entry)->string)
        {
          return (*entry)->refcount;
        }
      else
        {
          entry = &((*entry)->next);
        }
    }
  return 0;
}

void intstr_init()
{
  /// Initiate the global array of buckets
  buckets = calloc(BUCKETS, sizeof(entry_t* ));
}

void intstr_done()
{
  /// TODO
  for (int i=0; i<BUCKETS; ++i)
    {
      entry_t* cursor = buckets[i];
      while (cursor != NULL)
        {
          entry_t* next = cursor->next;
          free(cursor->string);
          free(cursor);
          cursor = next;
        }
    }
  free(buckets);
  free(NULL);   // ALLOWED (but also meaningless). (!)
}
