// Imperative and Object-Oriented Programming Methodology
// Uppsala University - 2018 Autumn
// Copyright © 2018 Henrik Schulze & Tobias Wrigstad
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include "interned.h"

#define BUCKETS 17  

typedef struct entry entry_t;

struct entry
{
  char* string;
  int refcount; 
  entry_t* next;
};

/// Global variable..
static entry_t** buckets; /// Denna kallas "strings" i 2.1.4 i uppgiftstexten.

/// Calculate hash for a string -- no need to read this function
static unsigned long string_hash(char* str)
{
  unsigned long result = 0;
  do
    {
      result = result*  31 +* str;
    }
  while (*++str != '\0'); 
  return result;
}

/// Helper function to create an entry in the hash table. Probably
/// no need to change it, but allowed.
static entry_t* entry_create(char* str, entry_t* next)
{
  entry_t* result = calloc(1,sizeof(entry_t));
  *result = (entry_t) {
    .string = strdup(str),
    .refcount = 1,
    .next = next
  };
  return result;
}

/// Returns the interned string for str. If str is already interned,
/// increase refcount by 1, else set initial refcount to 1.
char* intstr_create(char* str)
{
  /// Get the hash code for the string and use it to find the right bucket:
  unsigned long bucket = string_hash(str) % BUCKETS;
  entry_t* entry = buckets[bucket];
  /// Search through the bucket for the string:
  while (entry != NULL)
    {
      /// Use strcmp here -- because we expect str to be not interned
      if (strcmp(str, entry->string) == 0)
        {
          /// If the string was already interned, increase its refcount by 1
          /// and return the interned string:
          ++entry->refcount;
          return entry->string;
        }
      entry = entry->next; // string not found yet - iterate the cursor
    }
  /// First time called on str -- create a new entry, add it to
  /// the front of the bucket, and return the entry's string
  entry_t* new_entry = entry_create(str, buckets[bucket]);
  buckets[bucket] = new_entry;
  return new_entry->string;
}

/// Returns true if str is interned, else false.
bool intstr_is_interned(char* str)
{
  /// TODO
  if (str == NULL) return false;
  /// Get the hash code for the string and use it to find the right bucket:
  unsigned long bucket = string_hash(str) % BUCKETS;
  entry_t* cursor_ptr = buckets[bucket];
  /// Search through the bucket for a string:
  while (cursor_ptr != NULL)
    {
      /// Use strcmp here -- because we expect str to be not interned:
      if (strcmp(str, cursor_ptr->string) == 0)
        {
          /// If the string was already interned, return true:
          return true;
        }
      else
        {
          cursor_ptr = cursor_ptr->next;// string not found - iterate the cursor
        }
    }
  return false; /// string not found anywhere in the hash table
}

/// Decrease the refcount of str by 1. If refcount hits 0, remove the node
/// containing the string from the hash table. Also free the memory of both
/// the interned string and of the node.
void intstr_destroy(char* str)
{
  /// TODO
  if (str == NULL) return;
  unsigned long bucket = string_hash(str) % BUCKETS;
  entry_t** cursor_addr_of_ptr = &buckets[bucket];
  // ^^ ADDRESS of pointer to node. This is the cursor. ^^
  while ((*cursor_addr_of_ptr) != NULL) /// *cursor_addr_of_ptr = current node
    {
      if (str == (*cursor_addr_of_ptr)->string)
        {
        /// Decrease the refcount of str by 1:
          --((*cursor_addr_of_ptr)->refcount);
/// If refcount hits 0, remove the node containing the interned string
/// from the hash table structure. Then free the corresponding memory:
          if ((*cursor_addr_of_ptr)->refcount == 0)
            {
              entry_t *node_to_remove = *cursor_addr_of_ptr;
/// The following line is the trickiest code in this whole file!
/// - Replace pointer to current node with pointer to next node in linked list:
              *cursor_addr_of_ptr = (*cursor_addr_of_ptr)->next;
// ^^ Using a POINTER to node as cursor, this would be impossible! ^^
// ^^ This is NOT an iteration of the cursor. ^^
              free((node_to_remove)->string);
              free((node_to_remove));
            }
          return;
        }
      else
        {
        // str not found, iterate the cursor:
          cursor_addr_of_ptr = &((*cursor_addr_of_ptr)->next);
//^^ This IS iteration of the cursor. Change address to pointer of next node. ^^
        }
    }
}

/// Return the refcount for str if str is interned, else 0.
int intstr_refcount(char* str)
{
  if (str == NULL) return 0;

  unsigned long bucket = string_hash(str) % BUCKETS;
  entry_t** entry = &buckets[bucket];

  while ((*entry) != NULL)
    {
      if (str == (*entry)->string)
        {
          return (*entry)->refcount;
        }
      entry = &((*entry)->next);
    }

  return 0;
}

void intstr_init()
{
  /// Initiate the global array of buckets
  buckets = calloc(BUCKETS, sizeof(entry_t* ));
}

void intstr_done()
{
  /// TODO
  for(int i=0; i<BUCKETS; ++i)
    {
      entry_t* cursor_ptr = buckets[i];
      /// Search through the bucket to delete all entries:
      while (cursor_ptr != NULL)
        {
          entry_t* node_to_free = cursor_ptr;
          cursor_ptr = cursor_ptr->next; // iterate the cursor before freeing
          free(node_to_free->string);
          free(node_to_free);
        }
      buckets[i] = NULL;
      /// ^^ Otherwise the hash table contains nodes with unallocated memory! ^^
    }
  free(buckets);
}
